/* cfft_time.c - reading the time from raw Chilbolton radar files */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "chil.h"
#include "cfft.h"

/* Read the .raw file corresponding to a .fft file and get the start
 * time and ray spacing */

int
cfft_get_time(CFFTData *data, char *filename)
{
  ChilboltonFile chil_file;
  RasterInfo raster_info;
  Ray ray;
  char *rawname = strdup(filename);
  int len;
  unsigned short int raster;
  float utc1, utc2;

  data->starttime = -1.0;
  data->rayduration = 1.0 / 3600.0;
  data->day = 1;
  data->month = 1;
  data->year = 0;

  if (!rawname) {
    return CFFT_MEMORY_ERROR;
  }

  /* Look for a raw file */
  len = strlen(rawname);
  rawname[len-3] = 'r';
  rawname[len-2] = 'a';
  rawname[len-1] = 'w';
  
  if (CHILOpen(rawname, &chil_file) == FILE_CLOSED) {
    /* OK, look for a processed file instead */
    /* Move the tape number */
    rawname[len-12] = rawname[len-10];
    rawname[len-11] = rawname[len-9];
    rawname[len-10] = rawname[len-8];
    rawname[len-9] = rawname[len-7];
    /* Get raster number */
    raster = (rawname[len-6] - '0') * 10 + (rawname[len-5] - '0');
    /* Fill in the rest of the filename */
    sprintf(rawname + len - 8, "c94c.dat");
    if (CHILOpen(rawname, &chil_file) == FILE_CLOSED) {
      free(rawname);
      return CFFT_FILE_ERROR;
    }
    else {
      if (CHILSeekRaster(&chil_file, raster) == CHIL_ERROR) {
	free(rawname);
	CHILClose(&chil_file);
	return CFFT_FILE_ERROR;
      }
    }
  }

  if (CHILReadInformationRecord(&chil_file, &raster_info) == END_OF_DATA) {
    return CFFT_FILE_ERROR;
  }
  data->starttime = raster_info.hour + raster_info.minute / 60.0
    + raster_info.second / 3600.0;
  data->startday = raster_info.day;
  data->startmonth = raster_info.month;
  data->startyear = raster_info.year;

  CHILInitRay(&raster_info, 1, 1, &ray);

  if (CHILReadRay(&chil_file,&raster_info,&ray) == CHIL_OK) {
    utc1 = ray.utc;
    if (CHILReadRay(&chil_file,&raster_info,&ray) == CHIL_OK) {
      utc2 = ray.utc;
      if (utc2 > utc1) {
	data->rayduration = utc2-utc1;
      }
      else {
	data->rayduration = 24.0 + utc2-utc1;
      }
    }
  }
  
  CHILClose(&chil_file);
  CHILFreeRay(&ray);
  free(rawname);

  return CFFT_OK;
}

int
cfft_sync_time(CFFTData *data)
{
  if (data->starttime >= 0.0) {
    float newtime = data->starttime
      + data->profile * data->interval * data->rayduration;
    data->day = data->startday;
    data->month = data->startmonth;
    data->year = data->startyear;
    if (newtime > 24.0) {
      newtime -= 24.0;
      CHILIncrDate(&(data->day), &(data->month), &(data->year));
    }
    data->hour = (unsigned int) newtime;
    data->minute = (unsigned int) ((newtime - ((float) data->hour)) * 60.0);
    data->second = (unsigned int) ((newtime - ((float) data->hour)
			  - ((float) data->minute) / 60.0) * 3600.0);
  }
  return CFFT_OK;
}
