/* recalibrate.c -- recalibrate the fields in a NetCDF file */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <netcdf.h>

#include "nctools.h"
#include "chil.h"

static int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
      nc_strerror(ncstatus)); nc_close(ncid); exit(-1); \
} 
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s (%s)\n", \
	    __LINE__, nc_strerror(ncstatus), s); exit(-1); \
} 

static
char *
seek_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  if (nc_inq_atttype(ncid, varid, name, &type) != NC_NOERR
      || type != NC_CHAR) {
    return NULL;
  }

  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = malloc((length+1) * sizeof(char));
  if (!value) {
    fprintf(stderr, "Out of memory\n");
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}

static
short
get_short_attribute(int ncid, int varid, char *name)
{
  size_t length;
  short value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    fprintf(stderr, "Error: %s attribute is not of type short\n", name);
    exit(1);
  }
  if (length > 1) {
    fprintf(stderr, "Error: %s attribute has more than 1 element\n", name);
    exit(1);
  }
  handle_error(nc_get_att_short(ncid, varid, name, &value));
  return value;
}

static
void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.nc [file2.nc ...]\n", argv[0]);
  fprintf(stderr, "Options:\n"
	  "         -Zcal f  add f dB to  Z/Zh values\n"
	  "     -Zcalfile s  read Z/Zh calibration figures from file s\n"
	  "         -user s  name and email address of the user\n");

}

int
main(int argc, char **argv)
{
  char *user = NULL;
  float Zcal = 0;
  int iarg, ifile;
  char doZ = 0, usecalfile = 0;

  /* Parse switches */
  iarg=1;
  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg], "-Zcal") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg], "%f", &Zcal) < 1) {
	fprintf(stderr, "Error: failed to read Z calibration figure as a float\n");
	exit(-1);
      }
      else {
	doZ = 1;
      }
    }
    else if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	if (user) {
	  fprintf(stderr, "Warning: user specified more than once:"
		  " `%s' not used\n", user);
	}
	user = argv[++iarg];
      }
      else {
	fprintf(stderr, "Error: user not specified\n");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg],"-Zcalfile") == 0) {
      if (argc > iarg+1) {
	if (usecalfile) {
	  fprintf(stderr, "Error: calibration file specified more than once\n");
	  exit(1);
	}
	if (CHILReadCalibrationTable(argv[++iarg]) <= 0) {
	  fprintf(stderr, "Error reading %s as a calibration table\n",
		  argv[iarg]);
	  exit(1);
	}
	usecalfile = 1;
	doZ = 1;
      }
      else {
	fprintf(stderr, "Error: user not specified\n");
	exit(1);
      }
    }
    ++iarg;
  }
  if (iarg > argc-1) {
    Usage(argv);
    exit(0);
  }

  if (!doZ) {
    fprintf(stderr, "Warning: no action taken (no calibration figures supplied)\n");
    exit(0);
  }

  for (ifile = iarg; ifile < argc; ++ifile) {
    size_t ngates, nrays;
    size_t start[2] = {0, 0}, count[2] = {1, 1};
    float *Z = NULL;
    float nez;
    float missingval;
    int ncid;
    int timedimid, rangedimid;
    int Zid, nezid;
    int ray;
    char doZnow = doZ;
    char doneznow = 0;
    char domissingval = 0;

    /* Open file and get dimension sizes */
    handle_error(nc_open(argv[ifile], NC_WRITE | NC_NOCLOBBER, &ncid));
    handle_error(nc_redef(ncid));
    handle_error(nc_inq_dimid(ncid, "time", &timedimid));
    handle_error(nc_inq_dimlen(ncid, timedimid, &nrays));
    handle_error(nc_inq_dimid(ncid, "range", &rangedimid));
    handle_error(nc_inq_dimlen(ncid, rangedimid, &ngates));

    if (usecalfile) {
      int year = get_short_attribute(ncid, NC_GLOBAL, "year");
      int month = get_short_attribute(ncid, NC_GLOBAL, "month");
      int day = get_short_attribute(ncid, NC_GLOBAL, "day");
      if (CHILGetCalibrationValue(year, month, day, &Zcal)) {
	fprintf(stderr, "Warning: %s is out of the range of the calibration data\n",
		argv[ifile]);
	doZnow = 0;
      }
      else {
	/* Round to nearest 0.1 dB */
	Zcal = floor(Zcal*10.0 + 0.5) * 0.1;
      }
    }

    if (doZnow) {
      char *Zname = "Z";

      /* First see if the variable "Z" is present, then "Zh" */
      if (nc_inq_varid(ncid, Zname, &Zid) != NC_NOERR) {
	Zname = "Zh";
	if (nc_inq_varid(ncid, Zname, &Zid) != NC_NOERR) {
	  fprintf(stderr, "Warning: Z/Zh not found in %s\n", argv[ifile]);
	  doZnow = 0;
	}
      }
      if (doZnow) {
#define HISTORY_LENGTH 32
	char history[HISTORY_LENGTH];
	size_t oldZcallen = 0, missingvallen = 0;
	float oldZcal = 0.0, totalZcal = 0.0;

	/* Find out if there is a "missing value" */
	if (nc_inq_attlen(ncid, Zid, "missing_value", &missingvallen) == NC_NOERR) {
	  if (missingvallen == 1) {
	    nc_get_att_float(ncid, Zid, "missing_value", &missingval);
	    domissingval = 1;
	  }
	  else {
	    fprintf(stderr, "Warning: the \"missing_value\" attribute of %s"
		    " in %s has %d elements", Zname, argv[ifile], missingvallen);
	  }
	}

	/* Must also adjust the noise-equivalent reflectivity */
	if (nc_inq_varid(ncid, "nez", &nezid) == NC_NOERR) {
	  doneznow = 1;
	}

	/* Check what calibration has already been applied */
	/* First look for a "calibration_applied" local attribute */
	if (nc_inq_attlen(ncid, Zid, "calibration_applied", &oldZcallen) == NC_NOERR) {
	  if (oldZcallen == 1) {
	    nc_get_att_float(ncid, Zid, "calibration_applied", &oldZcal);
	  }
	  else {
	    fprintf(stderr, "Warning: the \"calibration_applied\" attribute of %s"
		    " in %s has %d elements", Zname, argv[ifile], oldZcallen);
	  }
	}
	else {
	  /* No "calibration_applied": look for global "command_line"
           * or "options" attributes */
	  char *optionsname = "command_line";
	  char *options = seek_char_attribute(ncid, NC_GLOBAL, optionsname);
	  if (!options) {
	    optionsname = "options";
	    options = seek_char_attribute(ncid, NC_GLOBAL, optionsname);
	    if (!options) {
	      fprintf(stderr, "Warning: no information about existing calibration"
		      " found in %s\n", argv[ifile]);
	    }
	  }
	  /* Search for a "-Zcal" argument */
	  if (options) {
	    char *ch = strstr(options, "-Zcal ");
	    if (ch) {
	      if (sscanf(ch+6, "%g", &oldZcal) < 1) {
		fprintf(stderr, "Warning: old calibration could not be determined"
			" from \"-Zcal\" argument of the \"%s\" global attribute\n",
			optionsname);
	      }
	    }
	    free(options);
	  }
	}

	/* Now we can start writing some attributes */
	/* First the "calibration_applied" attribute of Z */
	totalZcal = oldZcal + Zcal;
	nc_put_att_float(ncid, Zid, "calibration_applied", NC_FLOAT, 1, &totalZcal);

	/* Now the history and command line */
	snprintf(history, HISTORY_LENGTH, "Recalibrated (%+g dB)", Zcal);
	history[HISTORY_LENGTH-1] = '\0';
	handle_error(nct_add_history(ncid, history, user));
	handle_error(nct_add_command_line(ncid, argc, argv));

	/* Out of define mode */
	handle_error(nc_enddef(ncid));
	Z = malloc(ngates * sizeof(float));
	if (!Z) {
	  fprintf(stderr, "Out of memory\n");
	  exit(2);
	}
      }
    }
      
    /* Loop through the rays */
    for (ray = 0; ray < nrays; ray++) {
      start[0] = ray;
      count[1] = ngates;
      if (doZnow && Z) {
	int j;
	handle_error(nc_get_vara_float(ncid, Zid, start, count, Z));
	if (domissingval) {
	  for (j = 0; j < ngates; j++) {
	    if (Z[j] != missingval) {
	      Z[j] += Zcal;
	    }
	  }
	}
	else {
	  for (j = 0; j < ngates; j++) {
	    Z[j] += Zcal;
	  }
	}
	handle_error(nc_put_vara_float(ncid, Zid, start, count, Z));
	if (doneznow) {
	  handle_error(nc_get_var1_float(ncid, nezid, start, &nez));
	  nez += Zcal;
	  handle_error(nc_put_var1_float(ncid, nezid, start, &nez));
	}
      }
    }
    
    /* Finish up with current file */
    if (doZnow && Z) {
      free(Z);
    }
    nc_close(ncid);
  }
  exit(0);
}
